from rhetoric.view import view_config, view_defaults
from ..types import Language
from .forms import NewArticleForm


@view_defaults(route_name='articles.regional.index', renderer='json')
class ArticlesHandler(object):
    def __init__(self, request, language):
        self.request = request
        self.language = language
        self.region_strategy = Language.match(language)

    @view_config(request_method='GET')
    def show_local_entries(self):
        return {
            'language': self.language,
        }

    @view_config(request_method='POST', validate_form=NewArticleForm)
    def save_new_acrticle(self):
        return {'ok': True}

    @view_config(request_method='POST')
    def on_form_error(self):
        return {'ok': False, 'message': 'Form validation error.'}



import inspect
import functools


def viewdefaults(wrapped):
    """ This function is a copy of ``pyramid.util.viewdefaults``.

    Decorator for add_view-like methods which takes into account
    __view_defaults__ attached to view it is passed. Not a documented API but
    used by some external systems.
    """
    def wrapper(self, *arg, **kw):
        defaults = {}
        if arg:
            view = arg[0]
        else:
            view = kw.get('view')
        view = self.maybe_dotted(view)
        if inspect.isclass(view):
            defaults = getattr(view, '__view_defaults__', {}).copy()

        # We do not use action_method
        # Uncomment this when it's necessary
        # ----------------------------------
        #if not '_backframes' in kw:
        #    kw['_backframes'] = 1 # for action_method

        defaults.update(kw)
        return wrapped(self, *arg, **defaults)
    return functools.wraps(wrapped)(wrapper)



from django.http import HttpResponse
from django.middleware.csrf import CsrfViewMiddleware

from rhetoric.view import ViewCallback
from rhetoric.request import JsonBodyProperty


class CsrfProtectedViewDispatchMiddleware(CsrfViewMiddleware):

    def process_request(self, request):
        # We assume here that CsrfViewMiddleware doesn't have the process_request method
        # which should be called via super().
        # -------------------------------------------------
        # set request.response object as in
        # http://docs.pylonsproject.org/projects/pyramid/en/latest/api/request.html#pyramid.request.Request.response
        setattr(request, 'response', HttpResponse())
        setattr(request, 'json_body', JsonBodyProperty(request))


    def process_view(self, request, callback, callback_args, callback_kwargs):
        if isinstance(callback, ViewCallback):
            view_settings = callback.find_view_settings(request, callback_args, callback_kwargs)
            # Check the actual view callable rather than ViewCallback wrapper with CsrfViewMiddleware
            return super(CsrfProtectedViewDispatchMiddleware, self).process_view(
                request, view_settings['view'], callback_args, callback_kwargs
            )

        # The callable is a regular django view
        return super(CsrfProtectedViewDispatchMiddleware, self).process_view(
            request, callback, callback_args, callback_kwargs
        )



from django.http import HttpResponse
from rhetoric import view_config


@view_config(route_name='test.new.routes')
def default_view(request, param):
    return {}


@view_config(route_name='test.new.routes', request_method='POST')
def submit_form_view(request):
    return {}


@view_config(route_name='test.new.routes', request_method='POST', api_version='1.0')
def api_v1_submit_form_view(request):
    return {}


@view_config(route_name='blog.page', request_method='GET', renderer='json')
def blog_page(request, page_slug):
    # test custom response status api
    request.response.status_code = 201
    return {
        'page_slug': page_slug
    }


@view_config(route_name='blog.page', request_method='POST', renderer='json')
def blog_page_post(request, page_slug):
    return HttpResponse(page_slug)



import sys


def caller_module(level=2, sys=sys):
    """This function is taken from Pyramid Web Framework - ``pyramid.path.caller_module``."""
    module_globals = sys._getframe(level).f_globals
    module_name = module_globals.get('__name__') or '__main__'
    module = sys.modules[module_name]
    return module


def caller_package(level=2, caller_module=caller_module):
    """This function is taken from Pyramid Web Framework - ``pyramid.path.caller_package``."""
    # caller_module in arglist for tests
    module = caller_module(level+1)
    f = getattr(module, '__file__', '')
    if (('__init__.py' in f) or ('__init__$py' in f)): # empty at >>>
        # Module is a package
        return module
    # Go up one level to get package
    package_name = module.__name__.rsplit('.', 1)[0]
    return sys.modules[package_name]



import operator
from pkg_resources import parse_version


OPERATORS = {
    '>': operator.gt,
    '<': operator.lt,
    '==': operator.eq,
    '>=': operator.ge,
    '<=': operator.le
}


def match_api_version(request_version, allowed_version):
    """

    :param request_version:
    :param allowed_version: may be represented in following forms:
        1. ``VERSION``
        2. ``==VERSION`` (the same as above)
        3. ``>VERSION``
        4. ``<VERSION``
        5. ``>=VERSION``
        6. ``<=Version``
        7. Comma-separated list of 1-7 evaluated as AND
    :return: :raise ValueError:
    """
    distinct_versions = {version.strip() for version in allowed_version.split(',')}
    for distinct_version in distinct_versions:
        operation = OPERATORS.get(distinct_version[:2])
        if operation:
            # prepare cases #2, #5, #6
            compare_with = distinct_version[2:]
        else:
            operation = OPERATORS.get(distinct_version[0])
            if operation:
                # prepare cases #3, #4
                compare_with = distinct_version[1:]
            else:
                # prepare case #1
                compare_with = distinct_version
                operation = OPERATORS['==']

        # evaluate the case
        matched = operation(parse_version(request_version), parse_version(compare_with))
        if not matched:
            return False

    return True



from .models import Instruction


__all__ = ['filter_by_oid']


def filter_by_oid(instructions, oid):
    """ For a given list of instructions and an Order ID,
    return the list of instructions that reference that Order ID.

    :param instructions: list of instructions
    :type instructions: list or tuple
    :param oid: Order ID
    :type oid: int
    :return: list of instructions that reference ``oid``
    :rtype: list
    """
    # In python3, filter, map, zip, etc return an object which is iterable, but not a list.
    return list(filter(
        lambda i: Instruction.match(i)['filter_by_oid'](i, oid),
        instructions))


@Instruction.ORDER('filter_by_oid')
def filter_order_by_oid(order, oid):
    """

    :param order:
    :type order: :class:`tests.testapp.testapp.trading.models.Order`
    :param oid: Order ID
    :type oid: int
    """
    return order.tid == oid


@Instruction.CANCEL('filter_by_oid')
def filter_cancel_by_oid(cancel, oid):
    """

    :param cancel:
    :type cancel: :class:`tests.testapp.testapp.trading.models.Cancel`
    :param oid: Order ID
    :type oid: int
    """
    return cancel.xtid == oid

@Instruction.CANCEL_REPLACE('filter_by_oid')
def filter_cancel_replace_by_oid(cr, oid):
    """

    :param cr:
    :type cr: :class:`tests.testapp.testapp.trading.models.CancelReplace`
    :param oid:
    :return:
    """
    return cr.xr_tid == oid


# Alternative way
# ---------------

inline_matcher = Instruction.inline_match(
    ORDER = lambda o, oid: o.tid == oid,
    CANCEL = lambda c, oid: c.xtid == oid,
    CANCEL_REPLACE = lambda cr, oid: cr.xr_tid == oid
)

def filter_by_oid_alt(instructions, oid):
    """ For a given list of instructions and an Order ID,
    return the list of instructions that reference that Order ID.

    :param instructions: list of instructions
    :type instructions: list or tuple
    :param oid: Order ID
    :type oid: int
    :return: list of instructions that reference ``oid``
    :rtype: list
    """
    # In python3, filter, map, zip, etc return an object which is iterable, but not a list.
    return list(filter(
        lambda i: inline_matcher(i)(i, oid),
        instructions))



from django.http import HttpResponse
from django.core.serializers.json import DjangoJSONEncoder
from django.shortcuts import render


json_encode = DjangoJSONEncoder().encode


class JsonRendererFactory(object):
    def __init__(self, name):
        self.name = name

    def __call__(self, request, view_response):
        response = request.response
        response.content_type = 'application/json; charset=utf-8'
        response.content = json_encode(view_response)
        return response


class StringRendererFactory(object):
    def __init__(self, name):
        self.name = name

    def __call__(self, request, view_response):
        response = request.response
        response.content_type = 'text/plain; charset=utf-8'
        response.content = view_response
        return response


class DjangoTemplateRendererFactory(object):
    def __init__(self, name):
        self.name = name

    def __call__(self, request, context_dict):
        response = request.response
        httpresponse_kwargs = {
            'content_type': response['Content-Type'],
            'status': response.status_code
        }
        return render(request, self.name, context_dict, **httpresponse_kwargs)


BUILTIN_RENDERERS = {
    'json': JsonRendererFactory,
    'string': StringRendererFactory,
    '.html': DjangoTemplateRendererFactory,
}


class RenderingConfiguratorMixin(object):

    def add_renderer(self, name, factory):
        self.renderers[name] = factory

    def get_renderer(self, name):
        try:
            template_suffix = name.rindex(".")
        except ValueError:
            # period is not found
            renderer_name = name
        else:
            renderer_name = name[template_suffix:]

        try:
            return self.renderers[renderer_name](name)
        except KeyError:
            raise ValueError('No such renderer factory {}'.format(renderer_name))



"""
WSGI config for testapp project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/1.6/howto/deployment/wsgi/
"""
import os
import re


os.environ.setdefault("DJANGO_SETTINGS_MODULE", "testapp.settings")
# We must explicitly setup django here (starting from Django 1.7)
import django
if django.VERSION[:2] >= (1, 7):
    django.setup()

from .rhetoric_config import config

config.scan(
    ignore=[
    # do not scan settings modules
    re.compile('^testapp.testapp.settings[_]?[_a-z09]*$').match,
])

from django.core.wsgi import get_wsgi_application
application = get_wsgi_application()



from django.db import models

from ..types import Language


class AbstractRegionalArticle(models.Model):
    title = models.CharField(max_length=250, null=False, default='')
    text = models.TextField(null=False, default='')

    class Meta:
        abstract = True


@Language.ENGLISH('store:articles')
class EnglishArticle(AbstractRegionalArticle):
    pass

@Language.GERMAN('store:articles')
class GermanArticle(AbstractRegionalArticle):
    pass



from rhetoric.exceptions import ConfigurationError


class RoutesConfiguratorMixin(object):
    def add_route(self, name, pattern, rules=None, extra_kwargs=None):
        pattern = u'{}/{}'.format(self.route_prefix.rstrip('/'), pattern.lstrip('/'))
        self.routes[name] = {
            'name': name,
            'pattern': pattern,
            'rules': rules,
            'extra_kwargs': extra_kwargs,
            'viewlist': [],
        }

    def check_routes_consistency(self):
        for route_name, route in self.routes.items():
            viewlist = route['viewlist']
            if not viewlist:
                raise ConfigurationError(
                    'Route name "{name}" is not associated with a view callable.'.format(name=route_name)
                )
            for route_item in viewlist:
                if route_item.get('view') is None:
                    raise ConfigurationError(
                        'Route name "{name}" is not associated with a view callable.'.format(name=route_name)
                    )



import re

from django.conf.urls import patterns, include, url
from django.contrib import admin

from rhetoric import Configurator

from tests.testapp.testapp.blog import includeme as blog_config


urlpatterns = patterns('',
    # Examples:
    # url(r'^$', 'testapp.views.home', name='home'),
    # url(r'^blog/', include('blog.urls')),

    url(r'^admin/', include(admin.site.urls)),
)

# Rhetorical routing
# ------------------
from .rhetoric_config import config
urlpatterns.extend(config.django_urls())



from ..types import Language


def includeme(config):
    """
    :type config: :class:`rhetoric.Configurator`
    """
    RULES = {
        'language': Language
    }
    config.add_route('articles.regional.index', '/{language}', rules=RULES)



"""This module contains exception classes from Pyramid Web Framework."""

class ConfigurationError(Exception):
    """ Raised when inappropriate input values are supplied to an API
    method of a :term:`Configurator`"""



from django.views.decorators.http import require_http_methods
from django.views.decorators.http import require_GET, require_POST

from rhetoric import view_config
from rhetoric import view_defaults


def phony_decorator(view):
    """ Decorator for testing Rhetoric's decorator predicate.
    """
    def decorated(request, *args, **kwargs):
        return view(request, *args, **kwargs)
    return decorated


@view_config(route_name='index.dashboard', request_method='GET', renderer='index.html')
def dashboard(request):
    return {'method': 'GET'}


# request method restrictions using django decorators (for testing Rhetoric's "decorator" option)

@view_config(route_name='index.dashboard', request_method='POST', renderer='index.html',
             # only POST method will be handled, since the request_method param
             # is explicitly set
             decorator=require_http_methods(["GET", "POST"]))
def post_on_dashboard(request):
    request.response.status_code = 201
    return {'method': 'POST'}


@view_config(route_name='index.dashboard', request_method='PUT', renderer='index.html',
             # test multiple decorators combined
             decorator=(phony_decorator, phony_decorator))
def put_on_dashboard(request):
    return {'method': 'PUT'}


# Versioned views
# --------------------------------
@view_config(route_name='index.versions', request_method='GET', api_version='1.0', renderer='json')
def get_version_1(request):
    return {
        'version': '1.0'
    }

@view_config(route_name='index.versions', request_method='GET', api_version='>1.0, <2.0', renderer='json')
def get_version_1_range(request):
    return {
        'method': 'GET',
        'version': '>1.0, <2.0'
    }

@view_config(route_name='index.versions', request_method='POST', api_version='>1.0, <2.0', renderer='json')
def post_version_1_range(request):
    return {
        'method': 'POST',
        'version': '>1.0, <2.0'
    }

@view_config(route_name='index.versions', request_method='POST', api_version='>=2.0', renderer='json')
def post_version_2_range(request):
    return {
        'version': '>=2.0'
    }


@view_defaults(route_name='index.dashboard.api', renderer='json', api_version='>=1.0, <2.0')
class DashboardAPIv1(object):

    def __init__(self, request, *args, **kw):
        self.request = request

    @view_config(request_method='GET')
    def get_dashboard_items(self):
        return {
            'version': 1,
            'method': 'GET'
        }

    @view_config(request_method='POST')
    def post_dashboard_items(self):
        return {
            'version': 1,
            'method': 'POST'
        }


@view_defaults(route_name='index.dashboard.api', renderer='json', api_version='>=2.0')
class DashboardAPIv2(object):

    def __init__(self, request, *args, **kw):
        self.request = request

    @view_config(request_method='GET')
    def get_dashboard_items(self):
        return {
            'version': 2,
            'method': 'GET'
        }

    @view_config(request_method='POST')
    def post_dashboard_items(self):
        return {
            'version': 2,
            'method': 'POST'
        }


from django.core.urlresolvers import RegexURLPattern as DjangoRegexURLPattern
from django.http import HttpResponse
from django.http import Http404

import venusian

from rhetoric.predicates import match_api_version


class view_config(object):
    venusian = venusian

    def __init__(self, **settings):
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()
        depth = settings.pop('_depth', 0)

        def callback(scanner, name, obj):
            scanner.config.add_view(view=obj, **settings)

        info = self.venusian.attach(wrapped, callback, category='rhetoric', depth=depth + 1)
        if info.scope == 'class':
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__
        return wrapped


class view_defaults(view_config):
    """ This object is a copy of ``pyramid.view.view_defaults``.

    A class :term:`decorator` which, when applied to a class, will
    provide defaults for all view configurations that use the class. This
    decorator accepts all the arguments accepted by
    :meth:`pyramid.view.view_config`, and each has the same meaning.

    See :ref:`view_defaults` for more information.
    """

    def __call__(self, wrapped):
        wrapped.__view_defaults__ = self.__dict__.copy()
        return wrapped


class ViewCallback(object):
    """ Wrapper object around actual view callables that checks predicates during the request
    and processes results returned from view callables during the response.
    """
    def __init__(self, viewlist):
        self.viewlist = viewlist

    def __call__(self, request, *args, **kwargs):
        view_settings = self.find_view_settings(request, args, kwargs)
        response = view_settings['view'](request, *args, **kwargs)
        return self.process_callback_response(request, response, view_settings)

    def find_view_settings(self, request, args, kwargs):
        if hasattr(request, 'rhetoric_view_settings'):
            return getattr(request, 'rhetoric_view_settings')

        # cache
        for view_settings in self.viewlist:
            passed, request = self.check_predicates(view_settings, request, args, kwargs)
            if passed:
                setattr(request, 'rhetoric_view_settings', view_settings)
                return view_settings
        raise Http404

    def check_predicates(self, view_settings, request, req_args, req_kw):
        predicates = view_settings['predicates']

        # request method
        if request.method not in predicates['request_method']:
            return False, request

        # api version
        api_version_getter = view_settings['api_version_getter']
        if predicates['api_version'] and api_version_getter:
            request_api_version = api_version_getter(request)
            match = False
            for allowed_api_pattern in predicates['api_version']:
                match = match_api_version(request_api_version, allowed_api_pattern)
                if match:
                    break
            if not match:
                return False, request

        # form validation
        if predicates['validate_form']:
            if view_settings['form_data'] is None:
                form_data = request.POST
            else:
                form_data = view_settings['form_data'](request)
            form = predicates['validate_form'](form_data)

            setattr(request, 'form', form)
            if not form.is_valid():
                return False, request

        return True, request

    def process_callback_response(self, request, response, view_settings):
        if isinstance(response, HttpResponse):
            # Do not process standard django responses
            return response
        renderer = view_settings['renderer']
        return renderer(request, response)


class RegexURLPattern(DjangoRegexURLPattern):
    def __init__(self, regex, default_args=None, name=None, viewlist=None):
        super(RegexURLPattern, self).__init__(regex, ViewCallback(viewlist), default_args, name)
        self.viewlist = viewlist



import io
import os
import sys

from setuptools import find_packages
from setuptools import setup
from setuptools.command.test import test as TestCommand


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


long_description = read(
    os.path.join(os.path.dirname(__file__), 'README.rst'),
    os.path.join(os.path.dirname(__file__), 'CHANGES'),
)


# Additional Hooks
# ----------------------------
# Integrate py.test with setup.py:
# http://pytest.org/latest/goodpractises.html#integration-with-setuptools-test-commands

class PyTest(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


setup(
    name='Rhetoric',
    version='0.1.15',
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'Django>=1.4',
        'venusian>=1.0',
        'six',
    ],
    test_suite='tests',
    tests_require=['pytest', 'coverage'],
    package_data={
        # If any package contains *.txt or *.rst files, include them
        '': ['*.txt', '*.rst',]
    },
    include_package_data=True,
    cmdclass={
        'test': PyTest,
    },
    # PyPI metadata
    # Read more at http://docs.python.org/distutils/setupscript.html#meta-data
    author="Maxim Avanov",
    author_email="maxim.avanov@gmail.com",
    maintainer="Maxim Avanov",
    maintainer_email="maxim.avanov@gmail.com",
    description="Pyramid-like routes for Django projects",
    long_description=long_description,
    license="MIT",
    url="https://github.com/avanov/Rhetoric",
    download_url="https://github.com/avanov/Rhetoric",
    keywords="pyramid django routes",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Application Frameworks'
    ]
)



from django.db import models
from rhetoric.adt import adt


class Order(models.Model):
     tid = models.IntegerField(null=False)
     price = models.DecimalField(null=False, max_digits=16, decimal_places=4, default='0')
     size = models.IntegerField(null=False, default=0)


class Cancel(models.Model):
    xtid = models.IntegerField(null=False)


class CancelReplace(models.Model):
    xr_tid = models.IntegerField(null=False)
    new_price = models.DecimalField(null=False, max_digits=16, decimal_places=4, default='0')
    new_size = models.IntegerField(null=False, default=0)


class Instruction(adt):
    ORDER = Order
    CANCEL = Cancel
    CANCEL_REPLACE = CancelReplace



""" Request-related shortcuts.
"""
import json

from django.utils.functional import SimpleLazyObject

from .compat import text_


class JsonBodyProperty(SimpleLazyObject):
    def __init__(self, request):
        super(JsonBodyProperty, self).__init__(lambda: json.loads(text_(request.read(), request.encoding or 'utf-8')))



# Rhetorical routing
# ------------------
from rhetoric import Configurator
from tests.testapp.testapp.blog import includeme as blog_config


config = Configurator()
api_getter = lambda request: request.META['X-API-VERSION']
config.set_api_version_getter(api_getter)

config.include('tests.testapp.testapp.index')
config.include(blog_config, '/blog')
config.include('tests.testapp.testapp.articles', '/articles')



""" ADT stands for Algebraic data type
"""
from rhetoric.exceptions import ConfigurationError


class ADTConfiguratorMixin(object):

    def update_adt_registry(self, adt_meta):
        """

        :type adt_meta: dict
        """
        adt_type = adt_meta['type']
        self.adt[adt_type] = adt_meta

    def check_adt_consistency(self):
        for obj_id, adt_meta in self.adt.items():
            for case_name, case_meta in adt_meta['cases'].items():
                for variant, implementation in case_meta.items():
                    if implementation is None:
                        raise ConfigurationError(
                            'Case {case_name} of {type} is not exhaustive. '
                            'Here is the variant that is not matched: {variant} '
                            .format(
                                case_name=case_name,
                                type=str(adt_meta['type']),
                                variant=variant
                            )
                        )
        # All good. We no longer need the adt meta.
        delattr(self, 'adt')



import inspect

from rhetoric.exceptions import ConfigurationError
from rhetoric.util import viewdefaults


class ViewsConfiguratorMixin(object):

    @viewdefaults
    def add_view(self,
                 view=None,
                 route_name=None,
                 request_method=None,
                 attr=None,
                 api_version=None,
                 validate_form=None,
                 form_data=None,
                 decorator=None,
                 check_csrf=False,
                 renderer=None):
        """

        :param view: callable
        :param route_name:
        :type route_name: str or None
        :param request_method:
        :type request_method: str or tuple
        :param attr:
          This knob is most useful when the view definition is a class.

          The view machinery defaults to using the ``__call__`` method
          of the :term:`view callable` (or the function itself, if the
          view callable is a function) to obtain a response.  The
          ``attr`` value allows you to vary the method attribute used
          to obtain the response.  For example, if your view was a
          class, and the class has a method named ``index`` and you
          wanted to use this method instead of the class' ``__call__``
          method to return the response, you'd say ``attr="index"`` in the
          view configuration for the view.
        :type attr: str
        :param api_version:
        :type api_version: str or tuple
        :param validate_form: Django form class or None
        :param form_data: callable that accepts ``request`` and returns an object that will be passed
                          to ``validate_form``.
        :param decorator:
        :param check_csrf:
        :param renderer:
        :return: :raise ConfigurationError:
        """
        try:
            route = self.routes[route_name]
        except KeyError:
            raise ConfigurationError(
                'No route named {route_name} found for view registration'.format(route_name=route_name)
            )

        # Parse view
        # -----------------------------------------------
        if inspect.isclass(view):
            actual_method = attr if attr else '__call__'
            view = ClassViewWrapper(view, actual_method)

        # Add decorators
        # -----------------------------------------------
        def combine(*decorators):
            def decorated(view_callable):
                # reversed() is allows a more natural ordering in the api
                for decorator in reversed(decorators):
                    view_callable = decorator(view_callable)
                return view_callable
            return decorated

        if isinstance(decorator, tuple):
            decorator = combine(*decorator)

        if decorator:
            view = decorator(view)

        # csrf_exempt is used by Django CSRF Middleware
        # -----------------------------------------------
        view.csrf_exempt = not check_csrf

        # Register predicates
        # -------------------------------------
        if request_method is None:
            request_method = {'GET'}
        elif isinstance(request_method, str):
            request_method = {request_method}
        request_method = set(request_method)

        if validate_form:
            if len(request_method) > 1 and not form_data:
                raise ConfigurationError(
                    'You must explicitly specify the form_data parameter, because one of the '
                    'view handlers accepts multiple request methods: {route_name}'.format(route_name=route_name)
                )

            django_supported_form_methods = {'GET', 'POST'}
            target_methods = request_method.copy() | django_supported_form_methods
            unsupported_methods = target_methods ^ django_supported_form_methods
            if unsupported_methods and not form_data:
                raise ConfigurationError(
                    'You must explicitly specify the form_data parameter, because Django '
                    'does not create form dict for methods {methods}: {route_name}'.format(
                        methods=unsupported_methods,
                        route_name=route_name)
                )

        if api_version is not None:
            if isinstance(api_version, str):
                api_version = {api_version}
            api_version = set(api_version)

        # Renderers
        # -------------------------------------
        if renderer is None:
            renderer = 'string'

        # Save
        # -------------------------------------
        route_item = {
            'view': view,
            'attr': attr,
            'api_version_getter': self.api_version_getter,
            'form_data': form_data,
            'renderer': self.get_renderer(renderer),
            'predicates': {
                'request_method': request_method,
                'api_version': api_version,
                'validate_form': validate_form,
                },
        }
        route['viewlist'].append(route_item)


class ClassViewWrapper(object):
    def __init__(self, view_class, method_to_call):
        self.view_class = view_class
        self.method_to_call = method_to_call

    def __call__(self, request, *args, **kw):
        instance = self.view_class(request, *args, **kw)
        view = getattr(instance, self.method_to_call)
        return view()



from collections import OrderedDict
import inspect

import pkg_resources
import venusian

from rhetoric.config.rendering import RenderingConfiguratorMixin
from rhetoric.config.rendering import BUILTIN_RENDERERS
from rhetoric.config.routes import RoutesConfiguratorMixin
from rhetoric.config.views import ViewsConfiguratorMixin
from rhetoric.config.adt import ADTConfiguratorMixin
from rhetoric.path import caller_package
from rhetoric.exceptions import ConfigurationError
from rhetoric.view import view_config
from rhetoric.view import view_defaults
from rhetoric.url import route_path, create_django_route


__all__ = ['route_path', 'view_config', 'view_defaults', 'Configurator']


class Configurator(
    RoutesConfiguratorMixin,
    ViewsConfiguratorMixin,
    RenderingConfiguratorMixin,
    ADTConfiguratorMixin):

    venusian = venusian
    inspect = inspect

    def __init__(self, route_prefix=None, api_version_getter=None):
        if route_prefix is None:
            route_prefix = ''
        self.route_prefix = route_prefix
        self.api_version_getter = api_version_getter

        self.routes = OrderedDict()
        self.renderers = {}
        # ADT extension
        # -------------
        self.adt = {}

        self.setup_registry()

    def include(self, callable, route_prefix=None):
        if route_prefix is None:
            route_prefix = ''

        old_route_prefix = self.route_prefix
        route_prefix = u'{}/{}'.format(old_route_prefix.rstrip('/'), route_prefix.lstrip('/'))
        self.set_route_prefix(route_prefix)

        c = self.maybe_dotted(callable)
        module = self.inspect.getmodule(c)
        if module is c:
            try:
                c = getattr(module, 'includeme')
            except AttributeError:
                raise ConfigurationError(
                    "module {} has no attribute 'includeme'".format(module.__name__)
                )

        sourcefile = self.inspect.getsourcefile(c)

        if sourcefile is None:
            raise ConfigurationError(
                'No source file for module {} (.py file must exist, '
                'refusing to use orphan .pyc or .pyo file).'.format(module.__name__)
            )

        c(self)
        self.set_route_prefix(old_route_prefix)

    def django_urls(self):
        """Converts registered routes to a list of Django URLs"""
        return [
            create_django_route(
                name=route_dict['name'],
                pattern=route_dict['pattern'],
                rules=route_dict.get('rules'),
                extra_kwargs=route_dict.get('extra_kwargs'),
                viewlist=route_dict['viewlist']
            )
            for route_dict in self.routes.values()
        ]

    def scan(self, package=None, categories=None, onerror=None, ignore=None):
        if package is None:
            package = caller_package()

        package = self.maybe_dotted(package)

        scanner = self.venusian.Scanner(config=self)
        scanner.scan(package, categories=categories, onerror=onerror, ignore=ignore)
        self.check_routes_consistency()
        self.check_adt_consistency()

    def setup_registry(self):
        # Add default renderers
        # ---------------------
        for name, renderer in BUILTIN_RENDERERS.items():
            self.add_renderer(name, renderer)



    def maybe_dotted(self, dotted):
        if not isinstance(dotted, str):
            return dotted
        return self._pkg_resources_style(dotted)

    def _pkg_resources_style(self, value):
        """ This method is taken from Pyramid Web Framework.

        package.module:attr style
        """
        return pkg_resources.EntryPoint.parse('x={}'.format(value)).load(False)

    def set_route_prefix(self, prefix):
        self.route_prefix = prefix

    def set_api_version_getter(self, getter_callable):
        self.api_version_getter = getter_callable



import re
import venusian
import six


ADT_VARIANT_NAME_RE = re.compile('^[A-Z][0-9A-Z_]*$')


class ADTVariant(object):
    venusian = venusian

    def __init__(self, variant_of, name, value):
        """

        :param variant_of:
        :param name: variant name (uppercase variant name)
        :param value: variant value
        """
        self.variant_of = variant_of
        self.name = name
        self.value = value

    def is_primitive_type(self):
        if isinstance(self.value, six.string_types):
            primitive_type = True
        elif isinstance(self.value, six.integer_types):
            primitive_type = True
        else:
            primitive_type = False
        return primitive_type


    def __call__(self, case_name):
        """ This method is called when we declare variant cases. For instance:

            >>> class Language(adt):
            >>>     ENGLISH = 'en'
            >>>
            >>> @Language.ENGLISH('case_name')
            >>> def case_implementation():
            >>>     pass
            >>>

        :param case_name:
        :type case_name: str
        """
        cases_dict = self.variant_of.__adt__['cases']
        variants = self.variant_of.__adt__['variants']

        cases_dict.setdefault(case_name, {v: None for v in variants})

        def wrapper(wrapped):
            def callback(scanner, name, obj):
                """ This method will be called by venusian.
                The ``callback`` name is not reserved, we just pass this callable
                to ``self.venusian.attach()`` below. Although, callback's arguments list is strictly defined.
                """
                cases_dict = self.variant_of.__adt__['cases']
                case_implementations = cases_dict[case_name]
                if case_implementations[self.name] is not None:
                    # venusian may scan the same declarations multiple times during the app initialization,
                    # therefore we allow re-assignment of the same case implementations and prohibit
                    # any new implementations
                    if case_implementations[self.name] is not obj:
                        raise TypeError(
                            'Variant {variant} of {type} is already bound to the case {case} => {impl}. '
                            'Conflict at {target}'.format(
                                variant=self.name,
                                type=self.variant_of.__adt__['type'],
                                case=case_name,
                                impl=str(case_implementations[self.name]),
                                target=str(obj)
                            )
                        )
                case_implementations[self.name] = obj
                # Re-calculate matches
                self.variant_of.__adt__['matches'] = {
                    variant: {case: impl[variant] for case, impl in cases_dict.items()}
                    for variant in self.variant_of.__adt__['variants']
                }
                scanner.config.update_adt_registry(self.variant_of.__adt__)

            info = self.venusian.attach(wrapped, callback, category='rhetoric')
            return wrapped
        return wrapper


class ADTMeta(type):
    def __new__(mcs, class_name, bases, attrs):
        cls = type.__new__(mcs, class_name, bases, attrs)

        variants = {}
        for attr_name, value in attrs.items():
            if not ADT_VARIANT_NAME_RE.match(attr_name):
                continue
            variant = ADTVariant(variant_of=cls, name=attr_name, value=value)
            setattr(cls, attr_name, variant)
            variants[attr_name] = variant

        values = {attrs[variant_name]: variant_name for variant_name in variants}
        cls.__adt__ = {
            'type': str(cls),
            # set of adt variants
            'variants': variants,
            # dict of value => variant mappings
            'values': values,
            'cases': {},
            # dict of value => match instances.
            # Used by .match() for O(1) result retrieval
            'matches': {
                variant: {} for variant in variants
            }
        }
        return cls


@six.add_metaclass(ADTMeta)
class adt(object):
    __adt__ = {}

    class Mismatch(Exception):
        pass

    class PatternError(Exception):
        pass

    @classmethod
    def values(cls):
        return set(cls.__adt__['values'].keys())

    @classmethod
    def match(cls, value):
        """
        :rtype: dict or :class:`types.FunctionType`
        """
        variant = None
        for variant_name, variant_type in cls.__adt__['variants'].items():
            if variant_type.is_primitive_type():
                # We compare primitive types with equality matching
                if value == variant_type.value:
                    variant = variant_name
                    break
            else:
                # We compare non-primitive types with type checking
                if isinstance(value, variant_type.value):
                    variant = variant_name
                    break

        if variant is None:
            raise cls.Mismatch(
                u'Variant value "{value}" is not a part of the type {type}: {values}'.format(
                    value=value,
                    type=cls.__adt__['type'],
                    values=u', '.join(['{val} => {var}'.format(val=val, var=var)
                                       for val, var in cls.__adt__['values'].items()])
                )
            )

        return cls.__adt__['matches'][variant]

    @classmethod
    def inline_match(cls, **inline_cases):
        all_cases = set(cls.__adt__['variants'].keys())
        inline_cases = inline_cases.items()
        checked_cases = []
        for variant_name, fun in inline_cases:
            try:
                variant = cls.__adt__['variants'][variant_name]
            except KeyError:
                raise cls.PatternError(
                    'Variant {variant} does not belong to the type {type}'.format(
                        variant=str(variant_name),
                        type=cls.__adt__['type'],
                    )
                )
            all_cases.remove(variant.name)
            checked_cases.append((variant, fun))

        if all_cases:
            raise cls.PatternError(
                'Inline cases are not exhaustive.\n'
                'Here is the variant that is not matched: {variant} '.format(
                    variant=list(all_cases)[0]
                )
            )

        def matcher(value):
            for variant, fun in checked_cases:
                if variant.is_primitive_type():
                    if value == variant.value:
                        return fun
                else:
                    if isinstance(value, variant.value):
                        return fun

            raise cls.Mismatch(
                u'Variant value "{value}" is not a part of the type {type}: {values}'.format(
                    value=value,
                    type=cls.__adt__['type'],
                    values=u', '.join(['{val} => {var}'.format(val=val, var=var)
                                       for val, var in cls.__adt__['values'].items()])
                )
            )
        return matcher

    def __init__(self):
        raise TypeError('adt is a type, not an instance.')
